/*
Model 819 API include file.
4/2/2012

*/

#ifndef _S819_H_
#define _S819_H_

/*****************************************************
* CONSTANTS
*****************************************************/

//default and max bitrates for H.264
#define DEF_BR_QCIF             500000
#define DEF_BR_CIF              1000000
#define DEF_BR_D1               2000000

#define MIN_BR_MASTER           64000
#define MAX_BR_MASTER           4000000

#define MIN_BR_SUB              32000
#define MAX_BR_SUB              1000000

#define MIN_QP                  16
#define MAX_QP                  51

//Max GOP size
#define GOP_SIZE_MAX            256     //max GOP size

//Resolution constants
//Horizontal
#define HS_PAL_D1               720
#define HS_PAL_CIF              HS_PAL_D1/2
#define HS_PAL_QCIF             HS_PAL_D1/4

#define HS_NTSC_D1              720
#define HS_NTSC_CIF             HS_NTSC_D1/2
#define HS_NTSC_QCIF            HS_NTSC_D1/4

//Vertical
#define VS_PAL_D1               576
#define VS_PAL_CIF              VS_PAL_D1/2
#define VS_PAL_QCIF             VS_PAL_D1/4

#define VS_NTSC_D1              480
#define VS_NTSC_CIF             VS_NTSC_D1/2
#define VS_NTSC_QCIF            VS_NTSC_D1/4

#define MAX_FPS_JPG             6
#define MIN_FPS_JPG             1

//not supported
#define MIN_JPG_QUAL			0
#define MAX_JPG_QUAL			4

//Motion detection cells
#define HCELLS		16
#define VCELLS		12

/*****************************************************
* TYPES
*****************************************************/
//Supported resolutions
typedef enum {
    RES_QCIF = 0,
    RES_CIF,
    RES_D1,
    RES_MAX,           //for array sizing only
} S819_RESOLUTION;

//Supported TV standards
typedef enum {
    TVS_PAL = 0,
    TVS_NTSC,
    TVS_MAX,        //for array sizing
} S819_TVSTANDARD;

//Interlacing modes
typedef enum {
    MODE_INTERLACE = 0,
    MODE_PROGRESSIVE,
    MODE_MAX,           //for array sizing only
} S819_INTERLACE;

//Bitrate control modes
typedef enum {
    BITRATE_H264_NO_RC,             // Var with max limit
	BITRATE_H264_CBR,               //constant bitrate
	BITRATE_H264_VBR,               //variable bitrate
    BITRATE_MAX,        			//for array sizing only
} S819_BR_CTRL;

typedef enum {
    AUDIO_PCM,
	AUDIO_ALAW,
	AUDIO_ULAW,
	AUDIO_ADPCM,
    AUDIO_ADPCM_32K,
    AUDIO_TYPE_MAX,
} S819_AUDIOTYPE;


//Capture  mode (video decoder mode)
typedef struct {
    S819_TVSTANDARD     tvs;
    int                 bright;
    int                 contrast;
    int                 saturation;
    int                 hue;
    int                 sharpness;
} CHAN_MODE;

//Video Stream mode (compression)
typedef struct {
    S819_RESOLUTION     resolution;         //
    int                 fps;                //1..MAX (30 or 25)
    S819_INTERLACE      interlace;          //
    int                 gopSize;            //1..GOP_SIZE_MAX
    S819_BR_CTRL        brCtrl;             //CBR, VBR
    int                 bitrate;            //max bitrate (?)
} VIDEO_STREAM_MODE;

//uncompressed preview stream (not supported)
typedef struct {
	int					notSupported;
} PREVIEW_STREAM_MODE;

typedef enum {
	AUDIO_16BIT = 0,
	AUDIO_8BIT,
	AUDIO_SAMP_MAX,
} S819_AUDIOSAMPLE;

typedef enum {
    AUDIO_8000 = 0,
    AUDIO_16000,
    AUDIO_32000,
    AUDIO_44100,
    AUDIO_48000,
    AUDIO_RATE_MAX,
} S819_AUDIORATE;


typedef struct {
	S819_AUDIOTYPE		atype;
	S819_AUDIOSAMPLE	sampSize;
	S819_AUDIORATE		sampRate;
} AUDIO_STREAM_MODE;

typedef struct {
	int					sensitivity;
} ALARM_STREAM_MODE;

#define ALARM_SENS_MIN		0		//corresponds to max sensitivity
#define ALARM_SENS_MAX		15		//corresponds to min sensitivity

typedef struct {

} CAPTION_STREAM_MODE;

typedef union {
	VIDEO_STREAM_MODE	vMode;
	PREVIEW_STREAM_MODE pMode;
	AUDIO_STREAM_MODE	aMode;
	ALARM_STREAM_MODE	lMode;
	CAPTION_STREAM_MODE cMode;
} STREAM_MODE;

typedef struct {
    int                 ctrl;
	void				*aux;
} CONTROL;

//Channel structure
typedef struct {
    CHAN_MODE           cMode;          //channel capture mode
    STREAM_MODE         psMode;         //primary H.264 stream
    STREAM_MODE         ssMode;         //secondary H.264 stream
    STREAM_MODE         jsMode;         //JPEG stream
	STREAM_MODE			vsMode;			//preview stream
    STREAM_MODE         asMode;         //audio stream
	STREAM_MODE			lsMode;			//alarm stream
	STREAM_MODE			csMode;			//caption stream (in)
    CONTROL             ctrl;           //reserved
} S819_CHAN;

//Stream types and masks
typedef enum {
    STREAM_H264_PRI = 0,
    STREAM_H264_SEC,
    STREAM_JPEG,
    STREAM_YUV,
    STREAM_AUDIO,
    STREAM_ALARM,
	STREAM_CAPTION,
    STREAM_MAX
} S819_STREAM_TYPE;

#define SMASK(x)        (1<<(x))        //stream type mask

//Capture node handle.
//A capture node is a set of streams grouped together
//in a manner convenient for the application (using
//S819_CreateCnode fucntion).
//E.g. a capture node may consist of all primary
//streams for all channels, or of just one 
//particular stream.

typedef void *         HCNODE;


/*
//Crosspoint switch structure
typedef struct {
    

} S819_XPSW;
*/

typedef enum {
    H264_PSLICE  	= 1,
    H264_ISLICE  	= 5,
    H264_IDRSLICE 	= 6,
    H264_SPS        = 7,
    H264_PPS        = 8,
} SLICE_TYPE;

typedef enum {
    ALARM_VIDEO_LOST = 0x01,
    ALARM_NIGHT = 0x02,
    ALARM_BLIND = 0x04,
    ALARM_STDCHG = 0x08,
    ALARM_MV = 0x10,
} ALARM_TYPE;

typedef union {
	SLICE_TYPE			slice;
	ALARM_TYPE			atype;
} S819_SUBTYPE;

//length of MV data, bytes
#define MVBYTES		24

typedef struct {
    int                 chan;
    S819_STREAM_TYPE    stream;
	S819_SUBTYPE		subtype;
    char                *buf;
    int                 length;
    unsigned int        frm;
    unsigned int        pts;
    void                *rsv;
} BUFFER;


/*****************************************************
* ERROR CODES
*****************************************************/
typedef int         ECODE;                  //error code

#define ECODE_OK                0                   //success
#define ECODE_BASE              0                   //base for error codes

#define ECODE_SDK_INIT          (ECODE_BASE + 1)    //SDK initialization error
#define ECODE_SDK_START         (ECODE_BASE + 2)    //SDK start error
#define ECODE_SDK_COUNT         (ECODE_BASE + 3)    //error getting chip count
#define ECODE_NO_CHANS          (ECODE_BASE + 4)    //no 819 boards detected
#define ECODE_ILLEGAL_CHAN      (ECODE_BASE + 5)    //illegal channel number
#define ECODE_SETVD             (ECODE_BASE + 6)    //error setting VD
#define ECODE_CHAN_PARAM        (ECODE_BASE + 700)    //illegal S819_CHAN value
#define ECODE_SETVSP            (ECODE_BASE + 8)    //error setting VS for primary stream
#define ECODE_SETVSS            (ECODE_BASE + 9)    //error setting VS for secondary stream
#define ECODE_INVALID_HCNODE    (ECODE_BASE + 10)   //invalid HCNODE
#define ECODE_REG_CB            (ECODE_BASE + 11)   //error registering callback
#define ECODE_FN_UNDEFINED      (ECODE_BASE + 12)   //function undefined for stream
#define ECODE_INVALID_DATAP     (ECODE_BASE + 13)   //invalid data pointer
#define ECODE_INVALID_DATAB     (ECODE_BASE + 14)   //invalid data buffer
#define ECODE_DATA_TIMEOUT      (ECODE_BASE + 15)   //data timeout
#define ECODE_STRM_ATTACHED     (ECODE_BASE + 16)   //stream already attached to node
#define ECODE_STRM_NOTATTACHED  (ECODE_BASE + 17)   //stream not attached to node
#define ECODE_START_STRM        (ECODE_BASE + 18)   //error starting stream
#define ECODE_STOP_STRM         (ECODE_BASE + 19)   //error stopping stream
#define ECODE_SEMAPHORE         (ECODE_BASE + 20)   //semaphore error
#define ECODE_INVALID_PACKET    (ECODE_BASE + 21)   //invalid data packet
#define ECODE_INVALID_BUFFER    (ECODE_BASE + 22)   //invalid buffer pointer
#define ECODE_INVALID_BFR_DATA  (ECODE_BASE + 23)   //buffer contains invalid data
#define ECODE_RESOLUTION        (ECODE_BASE + 24)   //secondary resolution not lower than primary
#define ECODE_SETIBP            (ECODE_BASE + 25)   //error setting IBP parameters
#define ECODE_SETBRC            (ECODE_BASE + 26)   //error setting bitrate control
#define ECODE_SETBR             (ECODE_BASE + 27)   //error setting bitrate
#define ECODE_SETVP             (ECODE_BASE + 28)   //error setting decoder mode
#define ECODE_I2C               (ECODE_BASE + 29)   //I2C error
#define ECODE_SETJPEGFORMAT     (ECODE_BASE + 30)   //error setting JPEG format
#define ECODE_SETJPEGFPS        (ECODE_BASE + 31)   //error setting JPEG fps
#define ECODE_SETJPEGQUALITY    (ECODE_BASE + 32)   //error setting JPEG quality
#define ECODE_INVALID_STREAM    (ECODE_BASE + 33)   //invalid stream type
#define ECODE_ILLEGAL_CHAN_RCV  (ECODE_BASE + 34)   //illegal channel number received 
#define ECODE_THREAD0			(ECODE_BASE + 35)   //error creating thread

 
/*****************************************************
* API FUNCTIONS
*****************************************************/

//Detects the number of capture channels in the system
ECODE S819_Enumerate (int *nChans);

//Set channel operating mode.
//If chan = 0, s819chan is initialized with the defaults
ECODE S819_SetMode (int chan, S819_CHAN *s819chan);
//Creates a capture node
HCNODE S819_CreateCnode ();
void S819_DeleteCnode (HCNODE hCnode);
//Attaches channels/streams to a capture node
ECODE S819_AttachStreams (HCNODE hCnode, int chan, int streams);
//Start/stop selected streams
ECODE S819_StartStreams (HCNODE hCnode, int chan, int streams);
ECODE S819_StopStreams (HCNODE hCnode, int chan, int streams);
//Start/stop all streams of a CNODE
ECODE S819_StartAll (HCNODE hCnode);
ECODE S819_StopAll (HCNODE hCnode);

//Wait for data (blocking)
ECODE S819_WaitBuffer (HCNODE hCnode, BUFFER *buf);
ECODE S819_ReleaseBuffer (BUFFER *buf);

//crosspoint switch control
ECODE S819_SetXPSwitch (int board, int inp, int out);
ECODE S819_SetXPOut (int board, int mask);

ECODE S819_TOMStart (int *chanEn, int streams, int timeout_s);
void S819_TOMStop (void);

void S819_Close (void);

#endif  //_S819_H_


