/******************** (C) COPYRIGHT 2010 Intersil ********************
* File Name          : tc_chip.c
* Author             : zhang zhao
* Revision           : 1.2.1.1289
* Date               : 2012-04-05
* Description        : the tc_dev' staff for chip device
* Change log         : none
********************************************************************************
* THE PRESENT SOFTWARE IS DISTRIBUTED IN THE HOPE THAT IT WILL PROVIDE CUSTOMERS 
* WITH CODING INFORMATION TO SPEED UP THEIR PRODUCT DEVELOPMENT, BUT WITHOUT
* ANY WARRANTY; WITHOUT EVEN THE IMPLIED WARRANTY OF MERCHANTABILITY OR FITNESS
* FOR A PARTICULAR PURPOSE. AS A RESULT, TECHWELL SHALL NOT BE HELD LIABLE FOR ANY
* DIRECT, INDIRECT OR CONSEQUENTIAL DAMAGES WITH RESPECT TO ANY CLAIMS ARISING
* FROM THE CONTENT OF SUCH SOFTWARE AND/OR THE USE MADE BY CUSTOMERS OF THE
* CODING INFORMATION CONTAINED HEREIN IN CONNECTION WITH THEIR PRODUCTS.
*******************************************************************************/



#include <isil5864/tc_common.h>

/*******************************************************************************
* Function Name  : test_open
* Description    : open support function for chip device
* Input          : struct isil_chip_device *tcd,struct file * file
* Output         : NULL
* Return         : 0:success; !0:error
*******************************************************************************/
int test_open(struct isil_chip_device *tcd,struct file * file)
{
    int ret = 0;

    tc_trace;
    if(!tcd->ped || !tcd->ped->op)
        dbg("int %s have driver attached!\n",__FUNCTION__);

    ret = tcd->ped->op->open(tcd->ped);
    if(ret){
        ISIL_DBG(ISIL_DBG_INFO, ": driver open failed!\n");
        return ret;
    }
    file->private_data = tcd;	
    return ret;
}


/* after attach in open, so we can retrive our driver data struct from file->private_data
 * we must release all the resources by this chip dev and his chidren service dev's
 */
/*******************************************************************************
* Function Name  : test_release
* Description    : release support function for chip device
* Input          : struct isil_chip_device *chip_dev, struct file *file
* Output         : NULL
* Return         : 0:success; !0:error
*******************************************************************************/
int test_release(struct isil_chip_device *chip_dev, struct file *file)
{
    int ret;
    struct isil_chip_device *tcd = (struct isil_chip_device *)file->private_data;

    //printk("--->function %s called\n",__FUNCTION__);
    if(!tcd->ped || !tcd->ped->op){
        ISIL_DBG(ISIL_DBG_INFO, ": tcd->ped is NULL or ped->op is NULL\n");
        return -EINVAL;

    }
    ret = tcd->ped->op->close(tcd->ped);
    file->private_data = NULL;
    return ret;
}

/*******************************************************************************
* Function Name  : test_ioctl
* Description    : ioctl support function for chip device
* Input          : struct file * filp,unsigned int cmd, unsigned long arg
* Output         : NULL
* Return         : 0:success; !0:error
*******************************************************************************/
int test_ioctl(struct file * filp,unsigned int cmd, unsigned long arg)
{
    struct isil_chip_device *tcd = (struct isil_chip_device *)filp->private_data;
    int ret = 0;

    if(!tcd->ped || !tcd->ped->op){
        ISIL_DBG(ISIL_DBG_INFO, ": tcd->ped is NULL or ped->op is NULL\n");
        return -EINVAL;
    }

    down(&tcd->sem);
    ret = tcd->ped->op->ioctl(tcd->ped,cmd,arg);
    up(&tcd->sem);
    return ret;
}

struct tcd_file_operations  test_fop = {
    .open = test_open,
    .release = test_release,
    .ioctl = test_ioctl,
};






/*******************************************************************************
* Function Name  : tc_chip_init
* Description    : actully this is the function to creat and register the chip_device to tc_core
* Input          : struct isil_chip_device *tcd,unsigned long bus_id,unsigned long chip_id
* Output         : NULL
* Return         : 0:success; !0:error
*******************************************************************************/
int  tc_chip_init(struct isil_chip_device *tcd,unsigned long bus_id,unsigned long chip_id)
{
    int ret = 0;
    epkey_t key = {0};
    unsigned long idx = 0;

    dbg("@@@@:bus_id[%lu];chip_id[%lu]\n",bus_id,chip_id);
    for(;idx < 1; idx++){
		sprintf(tcd->name,"%s-%02lu","tw5864",chip_id);
		make_key((epkey_t *)&key,0x3f,0x3f,bus_id,0,chip_id,0);
		epobj_init(&tcd->epobj, ISIL, bus_id,TW5864,0,0,&key);
		tcd->tcd_fops = &test_fop;
		ret = register_chip_device(tcd);
		if(ret){
			ISIL_DBG(ISIL_DBG_INFO, ": regisrter chip dev failed!\n");
			return ret;
		}
	}
	
        return ret;
}




/*******************************************************************************
* Function Name  : tc_chip_exit
* Description    : unregister the chip_device
* Input          : struct isil_chip_device *tcd
* Output         : NULL
* Return         : NULL
*******************************************************************************/
void  tc_chip_exit(struct isil_chip_device *tcd)
{
    unregister_chip_device(tcd);
}





