/****************************************************************************************************************
* File Name         : isil_rc_wrapper.c
* Author            : shangchang Tang
* Revision          : 1.2.1.1275
* Date              : 2012-03-29
* Description       : this file focus on rc interface
* Change log        : 
*                   
****************************************************************************************************************
* THE PRESENT SOFTWARE IS DISTRIBUTED IN THE HOPE THAT IT WILL PROVIDE CUSTOMERS 
* WITH CODING INFORMATION TO SPEED UP THEIR PRODUCT DEVELOPMENT, BUT WITHOUT
* ANY WARRANTY; WITHOUT EVEN THE IMPLIED WARRANTY OF MERCHANTABILITY OR FITNESS
* FOR A PARTICULAR PURPOSE. AS A RESULT, TECHWELL SHALL NOT BE HELD LIABLE FOR ANY
* DIRECT, INDIRECT OR CONSEQUENTIAL DAMAGES WITH RESPECT TO ANY CLAIMS ARISING
* FROM THE CONTENT OF SUCH SOFTWARE AND/OR THE USE MADE BY CUSTOMERS OF THE
* CODING INFORMATION CONTAINED HEREIN IN CONNECTION WITH THEIR PRODUCTS.
****************************************************************************************************************/
#include    <isil5864/isil_common.h>


static isil_rc_driver_op    cbr_driver_op;
static isil_rc_driver_op    vbr_driver_op;
static isil_rc_driver_op    rc_driver_op;

/****************************************************************************************************************
* Function Name :   isil_cbr_rc_init
* Description   :   process cbr algo init.
* Input         :   param1 is h264 logic encode chan
* Output        :   null
* Return        :   null
****************************************************************************************************************/
static void isil_cbr_rc_init(isil_h264_logic_encode_chan_t *encoder)
{
    h264_encoder_t              *h264_rc_info;
    isil_h264_encode_property_t *encode_property;
    isil_h264_encode_control_t  *encode_control;
    int target_bitrate, target_fps, target_key_frame_intervals;

    h264_rc_info = &encoder->rc_driver.cbr.h264_rc_info;
    h264_rc_info->i_frame_type = X264_FRAME_TYPE_IDR;
    h264_rc_info->i_rc_frame_num = 0;
    h264_rc_info->channel_id = encoder->logic_chan_id;
    h264_rc_info->delta_time_stamp = 33;
    encode_property = &encoder->encode_property;
    encode_control = &encoder->encode_control;
    h264_rc_info->curr_qp = encode_property->encode_rc.op->get_qpi(&encode_property->encode_rc);
    target_bitrate = encode_property->op->get_target_bitrate(encode_property);
    target_fps = encode_property->op->get_target_fps(encode_property);
    target_key_frame_intervals = encode_property->op->get_keyFrameIntervals(encode_property);
    ISIL_rc_init(h264_rc_info, (encode_control->i_mb_x<<4), (encode_control->i_mb_y<<4), target_bitrate, target_fps, target_key_frame_intervals);
}

/****************************************************************************************************************
* Function Name :   isil264_cbr_update_rc
* Description   :   calc qp value based cbr algo
* Input         :   param1 is h264 logic encode chan
*                   param2 is curr encode frame's ptr
* Output        :   null
* Return        :   null
****************************************************************************************************************/
static void isil264_cbr_update_rc(isil_h264_logic_encode_chan_t	*encoder, isil_video_frame_tcb_t *video_frame)
{
    struct h264_encoder_t       *h264_rc_info;
    isil_h264_encode_control_t  *encode_control;

    encode_control = &encoder->encode_control;
    h264_rc_info = &encoder->rc_driver.cbr.h264_rc_info;
    h264_rc_info->curr_time_stamp = get_system_walltimer();
    if(h264_rc_info->i_rc_frame_num){
        if(h264_rc_info->curr_time_stamp >= h264_rc_info->last_time_stamp){
            h264_rc_info->delta_time_stamp = h264_rc_info->curr_time_stamp - h264_rc_info->last_time_stamp;
        }
    }
    h264_rc_info->last_time_stamp = h264_rc_info->curr_time_stamp; 
    if(video_frame->frame_type == H264_FRAME_TYPE_IDR){
        h264_rc_info->i_frame_type = X264_FRAME_TYPE_IDR;
    } else if(video_frame->frame_type == H264_FRAME_TYPE_I){
        h264_rc_info->i_frame_type = X264_FRAME_TYPE_I;
    } else {
        h264_rc_info->i_frame_type = X264_FRAME_TYPE_P;
    }
    ISIL264_update_rc(h264_rc_info);
    encode_control->i_curr_qp = h264_rc_info->curr_qp;
}

/****************************************************************************************************************
* Function Name :   isil264_cbr_update_rc_para
* Description   :   update cbr same param
* Input         :   param1 is h264 logic encode chan
* Output        :   null
* Return        :   null
****************************************************************************************************************/
static void  isil264_cbr_update_rc_para(isil_h264_logic_encode_chan_t *encoder)
{
    struct h264_encoder_t   *h264_rc_info;
    isil_chip_t             *chip;
    u32  bits_used_frame, non_zero_count, residual_bits_used_frame;

    h264_rc_info = &encoder->rc_driver.cbr.h264_rc_info;
    chip = encoder->chip;
    bits_used_frame = chip->io_op->chip_read32(chip, SLICE_TOTAL_BIT_NUMBER);
    non_zero_count = chip->io_op->chip_read32(chip, TOTAL_COEFF_NUMBER);
    residual_bits_used_frame = chip->io_op->chip_read32(chip, RES_TOTAL_BIT_NUMBER);
    ISIL264_update_rc_para(h264_rc_info, bits_used_frame, non_zero_count, residual_bits_used_frame);
    h264_rc_info->i_rc_frame_num++;
}

static void	isil264_cbr_jump_rc(isil_h264_logic_encode_chan_t *encoder)
{
}

static isil_rc_driver_op	cbr_driver_op =
{
    .init_rc = isil_cbr_rc_init,
    .set_qp = isil264_cbr_update_rc,
    .update_rc = isil264_cbr_update_rc_para,
    .jump_rc = isil264_cbr_jump_rc,
};

/****************************************************************************************************************
* Function Name :   isil_vbr_rc_init
* Description   :   process vbr algo init.
* Input         :   param1 is h264 logic encode chan
* Output        :   null
* Return        :   null
****************************************************************************************************************/
static void isil_vbr_rc_init(isil_h264_logic_encode_chan_t *encoder)
{
    struct h264_encoder_t       *h264_rc_info;
    isil_h264_encode_property_t *encode_property;
    isil_h264_encode_control_t  *encode_control;
    int target_bitrate, target_fps, target_key_frame_intervals;

    encode_property = &encoder->encode_property;
    encode_control = &encoder->encode_control;
    h264_rc_info = &encoder->rc_driver.vbr.h264_rc_info;
    h264_rc_info->i_frame_type = X264_FRAME_TYPE_IDR;
    h264_rc_info->i_rc_frame_num = 0;
    h264_rc_info->channel_id = encoder->logic_chan_id;
    h264_rc_info->delta_time_stamp = 33;
    encoder->rc_driver.vbr.image_level = encode_property->encode_rc.op->get_image_level(&encode_property->encode_rc);
    h264_rc_info->curr_qp = encode_property->encode_rc.op->get_qpi(&encode_property->encode_rc);
    target_bitrate = encode_property->op->get_target_bitrate(encode_property);
    target_fps = encode_property->op->get_target_fps(encode_property);
    target_key_frame_intervals = encode_property->op->get_keyFrameIntervals(encode_property);
    ISIL_rc_init_ext(h264_rc_info, (encode_control->i_mb_x<<4), (encode_control->i_mb_y<<4), target_bitrate, target_fps , target_key_frame_intervals, encoder->rc_driver.vbr.image_level);
}

/****************************************************************************************************************
* Function Name :   isil264_vbr_update_rc
* Description   :   calc qp value based vbr algo
* Input         :   param1 is h264 logic encode chan
*                   param2 is curr encode frame's ptr
* Output        :   null
* Return        :   null
****************************************************************************************************************/
static void isil264_vbr_update_rc(isil_h264_logic_encode_chan_t *encoder, isil_video_frame_tcb_t *video_frame)
{
    struct h264_encoder_t       *h264_rc_info;
    isil_h264_encode_control_t  *encode_control;

    encode_control = &encoder->encode_control;
    h264_rc_info = &encoder->rc_driver.vbr.h264_rc_info;
    h264_rc_info->curr_time_stamp = get_system_walltimer();
    if(h264_rc_info->i_rc_frame_num){
        if(h264_rc_info->curr_time_stamp >= h264_rc_info->last_time_stamp){
            h264_rc_info->delta_time_stamp = h264_rc_info->curr_time_stamp - h264_rc_info->last_time_stamp;
        }
    }
    h264_rc_info->last_time_stamp = h264_rc_info->curr_time_stamp; 
    if(video_frame->frame_type == H264_FRAME_TYPE_IDR){
        h264_rc_info->i_frame_type = X264_FRAME_TYPE_IDR;
    } else if(video_frame->frame_type == H264_FRAME_TYPE_I){
        h264_rc_info->i_frame_type = X264_FRAME_TYPE_I;
    } else {
        h264_rc_info->i_frame_type = X264_FRAME_TYPE_P;
    }
    ISIL264_update_rc_ext(h264_rc_info);
    encode_control->i_curr_qp = h264_rc_info->curr_qp;
}

/****************************************************************************************************************
* Function Name :   isil264_cbr_update_rc_para
* Description   :   update cbr same param
* Input         :   param1 is h264 logic encode chan
* Output        :   null
* Return        :   null
****************************************************************************************************************/
static void isil264_vbr_update_rc_para(isil_h264_logic_encode_chan_t *encoder)
{
    struct h264_encoder_t   *h264_rc_info;
    isil_chip_t             *chip;
    u32  bits_used_frame, non_zero_count, residual_bits_used_frame;

    h264_rc_info = &encoder->rc_driver.vbr.h264_rc_info;
    chip = encoder->chip;
    bits_used_frame = chip->io_op->chip_read32(chip, SLICE_TOTAL_BIT_NUMBER);
    non_zero_count = chip->io_op->chip_read32(chip, TOTAL_COEFF_NUMBER);
    residual_bits_used_frame = chip->io_op->chip_read32(chip, RES_TOTAL_BIT_NUMBER);
    ISIL264_update_rc_para_ext(h264_rc_info, bits_used_frame, non_zero_count, residual_bits_used_frame);
    h264_rc_info->i_rc_frame_num++;
}

static void isil264_vbr_jump_rc(isil_h264_logic_encode_chan_t *encoder)
{
}

static isil_rc_driver_op    vbr_driver_op =
{
    .init_rc = isil_vbr_rc_init,
    .set_qp = isil264_vbr_update_rc,
    .update_rc = isil264_vbr_update_rc_para,
    .jump_rc = isil264_vbr_jump_rc,
};

/****************************************************************************************************************
* Function Name :   isil_rc_init
* Description   :   process rc algo init.
* Input         :   param1 is h264 logic encode chan
* Output        :   null
* Return        :   null
****************************************************************************************************************/
static void isil_rc_init(isil_h264_logic_encode_chan_t *encoder)
{
    isil_rc_driver_t    *rc_driver;

    rc_driver = &encoder->rc_driver;
    rc_driver->rc_op = &rc_driver_op;
    rc_driver->cbr.rc_op = &cbr_driver_op;
    rc_driver->cbr.rc_op->init_rc(encoder);
    rc_driver->vbr.rc_op = &vbr_driver_op;
    rc_driver->vbr.rc_op->init_rc(encoder);
}

/****************************************************************************************************************
* Function Name :   isil264_update_rc
* Description   :   calc qp value
* Input         :   param1 is h264 logic encode chan
*                   param2 is curr encode frame's ptr
* Output        :   null
* Return        :   null
****************************************************************************************************************/
static void isil264_update_rc(isil_h264_logic_encode_chan_t *encoder, isil_video_frame_tcb_t *video_frame)
{
    isil_rc_driver_t	*rc_driver;
    isil_cbr_driver_t	*cbr_driver;
    isil_vbr_driver_t	*vbr_driver;	
    isil_h264_encode_property_t	*encode_property;
    isil_h264_encode_control_t	*encode_control;
    isil_h264_encode_rc_t       *encode_rc;

    rc_driver = &encoder->rc_driver;
    encode_property = &encoder->encode_property;
    encode_rc      = &encode_property->encode_rc;
    encode_control = &encoder->encode_control;
    switch(encode_rc->op->get_rc_type(encode_rc)){
        case ISIL_H264_CBR:
            cbr_driver = &rc_driver->cbr;
            if(cbr_driver->rc_op->set_qp != NULL){
                cbr_driver->rc_op->set_qp(encoder, video_frame);
            } else {
                encode_control->i_curr_qp = encode_rc->op->get_qpi(encode_rc);
            }
            break;
        case ISIL_H264_VBR:
            vbr_driver = &rc_driver->vbr;
            if(vbr_driver->rc_op->set_qp != NULL){
                vbr_driver->rc_op->set_qp(encoder, video_frame);
            } else {
                encode_control->i_curr_qp = encode_rc->op->get_qpi(encode_rc);
            }
            break;
        default:
            if((video_frame->frame_type==H264_FRAME_TYPE_IDR) || (video_frame->frame_type==H264_FRAME_TYPE_I)){
                encode_control->i_curr_qp = encode_rc->op->get_qpi(encode_rc);
            } else if(video_frame->frame_type==H264_FRAME_TYPE_P){
                encode_control->i_curr_qp = encode_rc->op->get_qpp(encode_rc);
            } else {
                encode_control->i_curr_qp = encode_rc->op->get_qpb(encode_rc);
            }
            break;
    }
}

static void isil264_jump_rc(isil_h264_logic_encode_chan_t *encoder)
{
    isil_rc_driver_t	*rc_driver;
    isil_cbr_driver_t	*cbr_driver;
    isil_vbr_driver_t	*vbr_driver;	

    rc_driver = &encoder->rc_driver;
    cbr_driver = &rc_driver->cbr;
    vbr_driver = &rc_driver->vbr;
    cbr_driver->rc_op->jump_rc(encoder);
    vbr_driver->rc_op->jump_rc(encoder);
}

/****************************************************************************************************************
* Function Name :   isil264_update_rc_para
* Description   :   update rc same param
* Input         :   param1 is h264 logic encode chan
* Output        :   null
* Return        :   null
****************************************************************************************************************/
static void  isil264_update_rc_para(isil_h264_logic_encode_chan_t *encoder)
{
    isil_rc_driver_t	*rc_driver;
    isil_cbr_driver_t	*cbr_driver;
    isil_vbr_driver_t	*vbr_driver;	
    isil_h264_encode_property_t	*encode_property;
    isil_h264_encode_control_t	*encode_control;
    isil_h264_encode_rc_t       *encode_rc;

    rc_driver = &encoder->rc_driver;
    encode_property = &encoder->encode_property;
    encode_rc      = &encode_property->encode_rc;
    encode_control = &encoder->encode_control;
    switch(encode_rc->op->get_rc_type(encode_rc)){
        case ISIL_H264_CBR:
            cbr_driver = &rc_driver->cbr;
            if(cbr_driver->rc_op->update_rc != NULL){
                cbr_driver->rc_op->update_rc(encoder);
            } else {
                cbr_driver->rc_op->jump_rc(encoder);
            }
            break;
        case ISIL_H264_VBR:
            vbr_driver = &rc_driver->vbr;
            if(vbr_driver->rc_op->update_rc != NULL){
                vbr_driver->rc_op->update_rc(encoder);
            } else {
                vbr_driver->rc_op->jump_rc(encoder);
            }
            break;
         default:
            rc_driver->rc_op->jump_rc(encoder);
            break;
    }
}

static isil_rc_driver_op    rc_driver_op =
{
    .init_rc = isil_rc_init,
    .set_qp = isil264_update_rc,
    .update_rc = isil264_update_rc_para,
    .jump_rc = isil264_jump_rc,
};

/****************************************************************************************************************
* Function Name :   init_rc_driver
* Description   :   init rc module.
* Input         :   param1 is h264 logic encode chan
* Output        :   null
* Return        :   null
****************************************************************************************************************/
void    init_rc_driver(isil_h264_logic_encode_chan_t *h264_logic_encode_chan)
{
    if(h264_logic_encode_chan != NULL){
        isil_rc_driver_t    *rc_driver = &h264_logic_encode_chan->rc_driver;
        rc_driver->rc_op = &rc_driver_op;
        rc_driver->rc_op->init_rc(h264_logic_encode_chan);
    }
}


